package com.jiuqi.util;

import java.util.ArrayList;

public class SequencialList
{
    private static final int NODE_SIZE = 1024;
    private static final int NODE_SHIFT = 10;
    private static final int NODE_OFFSET = 1023;
    private int size;
    private Object[] firstNode;
    private Object[] lastNode;
    private Object[] cacheNode;
    private int cachePage;
    
    public SequencialList() {
        this.clear();
    }
    
    public void clear() {
        this.size = 0;
        while (this.firstNode != null) {
            this.lastNode = (Object[])this.firstNode[1024];
            this.firstNode[1024] = null;
            this.firstNode = this.lastNode;
        }
        this.lastNode = null;
        this.cacheNode = null;
    }
    
    public void add(final Object value) {
        final int offset = this.size & 0x3FF;
        if (offset == 0) {
            final Object[] newNode = new Object[1025];
            if (this.lastNode != null) {
                this.lastNode[1024] = newNode;
                this.lastNode = newNode;
            }
            else {
                this.firstNode = newNode;
                this.lastNode = newNode;
            }
        }
        this.lastNode[offset] = value;
        ++this.size;
    }
    
    public int size() {
        return this.size;
    }
    
    public Object get(int index) {
        if (index < 0 || index >= this.size) {
            return null;
        }
        if (this.cacheNode != null) {
            final int page = index >> 10;
            if (page == this.cachePage) {
                index &= 0x3FF;
                return this.cacheNode[index];
            }
            if (page > this.cachePage) {
                index -= this.cachePage << 10;
            }
            else {
                this.cachePage = 0;
                this.cacheNode = this.firstNode;
            }
        }
        else {
            this.cachePage = 0;
            this.cacheNode = this.firstNode;
        }
        while (index >= 1024) {
            index -= 1024;
            this.cacheNode = (Object[])this.cacheNode[1024];
            ++this.cachePage;
        }
        return this.cacheNode[index];
    }
    
    public void set(int index, final Object value) {
        if (index < 0 || index >= this.size) {
            return;
        }
        if (this.cacheNode != null) {
            final int page = index >> 10;
            if (page == this.cachePage) {
                index &= 0x3FF;
                this.cacheNode[index] = value;
                return;
            }
            if (page > this.cachePage) {
                index -= this.cachePage << 10;
            }
            else {
                this.cachePage = 0;
                this.cacheNode = this.firstNode;
            }
        }
        else {
            this.cachePage = 0;
            this.cacheNode = this.firstNode;
        }
        while (index >= 1024) {
            index -= 1024;
            this.cacheNode = (Object[])this.cacheNode[1024];
            ++this.cachePage;
        }
        this.cacheNode[index] = value;
    }
    
    public int indexOf(final Object data) {
        int result = 0;
        if (data != null) {
            for (Object[] node = this.firstNode; result < this.size; result += 1024, node = (Object[])node[1024]) {
                for (int count = (this.size - result >= 1024) ? 1024 : (this.size - result), i = 0; i < count; ++i) {
                    if (data.equals(node[i])) {
                        return result + i;
                    }
                }
            }
        }
        else {
            for (Object[] node = this.firstNode; result < this.size; result += 1024, node = (Object[])node[1024]) {
                for (int count = (this.size - result >= 1024) ? 1024 : (this.size - result), i = 0; i < count; ++i) {
                    if (node[i] == null) {
                        return result + i;
                    }
                }
            }
        }
        return -1;
    }
    
    public int remove(final Object data) {
        final int index = this.indexOf(data);
        if (index >= 0) {
            this.remove(index);
        }
        return index;
    }
    
    public Object remove(int index) {
        if (index < 0 || index >= this.size) {
            return null;
        }
        int tailCount = this.size - index - 1;
        if (this.cacheNode != null) {
            final int page = index >> 10;
            if (page == this.cachePage) {
                index &= 0x3FF;
            }
            else if (page > this.cachePage) {
                index -= this.cachePage << 10;
            }
            else {
                this.cachePage = 0;
                this.cacheNode = this.firstNode;
            }
        }
        else {
            this.cachePage = 0;
            this.cacheNode = this.firstNode;
        }
        while (index >= 1024) {
            index -= 1024;
            this.cacheNode = (Object[])this.cacheNode[1024];
            ++this.cachePage;
        }
        final Object result = this.cacheNode[index];
        Object[] node = this.cacheNode;
        while (tailCount > 0) {
            int count = index + tailCount;
            if (count > 1024) {
                count = 1024;
            }
            for (int i = index; i < count; ++i) {
                node[i] = node[i + 1];
            }
            tailCount = tailCount - count + index;
            if (tailCount > 0) {
                index = 0;
                final Object[] nextNode = (Object[])node[1024];
                node[1023] = nextNode[0];
                node = nextNode;
            }
        }
        --this.size;
        if ((this.size & 0x3FF) == 0x0) {
            (this.lastNode = node)[1024] = null;
        }
        return result;
    }
    
    public ArrayList toArrayList() {
        final ArrayList result = new ArrayList(this.size);
        for (int i = 0; i < this.size; ++i) {
            result.add(this.get(i));
        }
        return result;
    }
    
    public static void test1() {
        final int count = 1000000;
        final ArrayList list1 = new ArrayList();
        final SequencialList list2 = new SequencialList();
        long curTime = System.currentTimeMillis();
        for (int i = 0; i < count; ++i) {
            list1.add(new Integer(i));
        }
        System.out.print("ArrayList Create: ");
        System.out.println(System.currentTimeMillis() - curTime);
        curTime = System.currentTimeMillis();
        for (int i = 0; i < count; ++i) {
            list2.add(new Integer(i));
        }
        System.out.print("SequencialList Create: ");
        System.out.println(System.currentTimeMillis() - curTime);
        curTime = System.currentTimeMillis();
        for (int k = 0; k < 10; ++k) {
            for (int j = 0; j < count; ++j) {
                if (list1.get(j) != j) {
                    System.out.println("Error");
                }
            }
        }
        System.out.print("ArrayList Get: ");
        System.out.println(System.currentTimeMillis() - curTime);
        curTime = System.currentTimeMillis();
        for (int k = 0; k < 10; ++k) {
            for (int j = 0; j < count; ++j) {
                if ((int)list2.get(j) != j) {
                    System.out.println("Error");
                }
            }
        }
        System.out.print("SequencialList Get: ");
        System.out.println(System.currentTimeMillis() - curTime);
        curTime = System.currentTimeMillis();
        for (int i = 0; i < count; i += 3000) {
            if (list1.indexOf(list1.get(i)) != i) {
                System.out.println("Error");
            }
        }
        System.out.print("ArrayList indexOf: ");
        System.out.println(System.currentTimeMillis() - curTime);
        curTime = System.currentTimeMillis();
        for (int i = 0; i < count; i += 3000) {
            if (list2.indexOf(list2.get(i)) != i) {
                System.out.println("Error");
            }
        }
        System.out.print("SequencialList indexOf: ");
        System.out.println(System.currentTimeMillis() - curTime);
        curTime = System.currentTimeMillis();
        for (int i = 0; i < count; i += 1000) {
            list1.remove(i);
        }
        System.out.print("ArrayList remove: ");
        System.out.println(System.currentTimeMillis() - curTime);
        curTime = System.currentTimeMillis();
        for (int i = 0; i < count; i += 1000) {
            list2.remove(i);
        }
        System.out.print("SequencialList remove: ");
        System.out.println(System.currentTimeMillis() - curTime);
    }
    
    public static void test2() {
        final SequencialList list = new SequencialList();
        for (int i = 0; i < 4000; ++i) {
            list.add(String.valueOf(i));
        }
        for (int i = 0; i < 4000; ++i) {
            list.get(i);
        }
        for (int i = 0; i < 4000; ++i) {
            list.set(i * 2, null);
        }
        for (int i = 400; i < 4000; ++i) {
            System.out.println(list.get(i));
        }
    }
    
    public static void main(final String[] args) {
        test1();
    }
}
